//---------------------------------------------------------------------------

#pragma hdrstop

#include "ArmScope.h"
#include "armscope_cmd.h"
#include "ThreadComm.h"
#include "usb.h"
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <algorithm>
#include <assert.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "libusb.lib"

ArmScope armScope;
struct usb_device *my_dev = NULL;
struct usb_dev_handle *usb_handle = NULL;

#define VENDOR_ID   0x03EB     
#define PRODUCT_ID  0x6130


static struct usb_device *device_init(void)
{
    struct usb_bus *usb_bus;
    struct usb_device *dev;
    usb_init();
//    armScope.log("USB: Enumerating devices...");
    usb_find_busses();
    usb_find_devices();
    for (usb_bus = usb_get_busses(); usb_bus;
            usb_bus = usb_bus->next)
    {
        for (dev = usb_bus->devices; dev;
                dev = dev->next)
        {
            if ((dev->descriptor.idVendor == VENDOR_ID) && (dev->descriptor.idProduct == PRODUCT_ID))
            {
//                    Log("Device found\r\n");
                return dev;
            }
        }
    }
    return NULL;
}


ArmScope::ArmScope():
	threadComm(NULL),
	hdr_bytes(0),
	hdr(NULL)
{
    bDeviceOpened = false;
	bStopRequest = false;
	callbackTrigger = NULL;
    callbackData = NULL;
    callbackLog = NULL;
	callbackConnect = NULL;

	// some default values
	settings.iChannelCount = 0; // this will allow to ignore setting array pointers inside settings structure to valid values
	settings.iSamplingPeriod = 1;
	settings.eTriggerType = TRIGGER_MANUAL;
	settings.iTriggerLevel = 0;
	settings.iBufferSize = 4096;

	upcoming_settings = settings;
	
	frame_position = 0;
	triggered = false;
	bManualTrigger = false;
}

ArmScope::~ArmScope()
{
}

void ArmScope::log(char *lpData, ...)
{
    static char buf[256]; //determines max message length
    va_list ap;
    va_start(ap, lpData);
    vsnprintf(buf, sizeof(buf), lpData, ap);
    va_end(ap);
    buf[sizeof(buf)-1] = 0;

    if (callbackLog)
        callbackLog(buf);
}

int ArmScope::Connect(void)
{
	assert(callbackTrigger && callbackData && callbackLog && callbackConnect);

	if (conn_state != DEVICE_DISCONNECTED)
		Disconnect();

	my_dev = device_init();
    if (my_dev == NULL)
    {
		SetConnectionState(DEVICE_DISCONNECTED, "Device not found. Connect device and restart program.");
		return 1;
    }
    else
    {
		//open device
        usb_handle = usb_open( my_dev );
        if (usb_handle == NULL)
        {
			SetConnectionState(DEVICE_DISCONNECTED, "Error opening device handle.");
            return 2;
        }
        else
        {
            //set configuration
            int ret;
            if ( ( ret = usb_set_configuration( usb_handle, 1 ) ) < 0 )
            {
				SetConnectionState(DEVICE_DISCONNECTED, "Failed to set device interface configuration.");
				log("usb config set failed with status %d", ret);
				return 3;
            }
            if ( ( ret = usb_claim_interface( usb_handle, 1 ) ) < 0 )   //chyba interfejs 1 - ten od BULK
            {
				SetConnectionState(DEVICE_DISCONNECTED, "Failed to claim USB interface.");
				Disconnect();
                log("FAIL: %s", usb_strerror());
                return 4;
            }
            else
            {
				threadComm = new ThreadComm(false);
				threadComm->callbackLog = callbackLog;
				threadComm->callbackDataRcv = DataRcvCb;
				threadComm->callbackConnect = callbackConnect;
				threadComm->Execute();

				SetConnectionState(DEVICE_CONNECTING, "Connecting...");
				if (threadComm->SendCmd(CMD_FUNCTION, CMD_FUNCTION_IDENTIFY) == 0)
				{
					log("Waiting for ID reply");
				}
				else
				{
					return 5;
				}
            }
        }
    }
    return 0;
}

int ArmScope::Disconnect(void)
{
	SetConnectionState(DEVICE_DISCONNECTING, "Disconnecting...");
	if (threadComm)
	{
		threadComm->connected = false;
		//threadComm->SetContinuousTrigger(false);
		Sleep(600);
		delete threadComm;
		threadComm = NULL;
	}
	if (usb_handle == NULL)
        return 0;
    usb_close(usb_handle);
	usb_handle = NULL;
	SetConnectionState(DEVICE_DISCONNECTED, "Disconnected");
	hdr_bytes = 0;
	hdr = NULL;
	return 0;
}

int ArmScope::ManualTrigger(void)
{
	if (!threadComm)
		return -1;
	if (conn_state != DEVICE_CONNECTED)
		return -2;
	bManualTrigger = true;
	return 0;
}

int ArmScope::SetTriggerType(enum E_TRIGGER_TYPE type)
{
	if (!threadComm)
		return -1;
	if (type == TRIGGER_MANUAL)
	{
		upcoming_settings.eTriggerType = type;
	}
	else if (type == TRIGGER_SLOPE_UP)
	{
		upcoming_settings.eTriggerType = type;
	}
	else if (type == TRIGGER_SLOPE_DOWN)
	{
		upcoming_settings.eTriggerType = type;
	}
	else if (type == TRIGGER_CONTINUOUS)
	{
		upcoming_settings.eTriggerType = type;
	}
	else
		return -2;
	return 0;
}

int ArmScope::SetCouplingType(enum E_COUPLING_TYPE type)
{
	if (!threadComm)
		return -1;
	enum E_SUBCOMMAND_FUNCTION_TYPE cmd;
	if (type == COUPLING_DC)
		cmd = CMD_FUNCTION_COUPLING_DC;
	else if (type == COUPLING_AC)
		cmd = CMD_FUNCTION_COUPLING_AC;
	else
		return -2;
	return threadComm->SendCmd(CMD_FUNCTION, cmd);
}

int ArmScope::SetBufferSize(int size)
{
	if (!threadComm)
		return -1;
	upcoming_settings.iBufferSize = size;
	return 0;
}

int ArmScope::SetSensitivity(int id)
{
	if (id < 0 || id > 5)
		return -1;
	return threadComm->SendCmd(CMD_SET_SENSITIVITY, id + 1);
}

int ArmScope::SetSampling(int id)
{
	if (id < 0 || id > 10)
		return -1;
	upcoming_settings.iSamplingPeriod = (1 << id);
	return 0;
}

int ArmScope::SetTriggerLevel(int level)
{
	if (level < -100 || level > 100)
		return -1;
	upcoming_settings.iTriggerLevel = level;
	return 0;
}

static unsigned char* memsearch(const unsigned char* hay, int haylen, const unsigned char* needle, int needlelen)
{
	for (int i=0; i<=haylen-needlelen; i++)
	{
		if (memcmp(hay+i, needle, needlelen) == 0)
			return (unsigned char*)(hay+i);
	}
	return NULL;
}

void ArmScope::DataRcvCb(unsigned char data[], unsigned int len)
{
	//log("DataRcv %d B", len);
	//ConnectCb(DEVICE_CONNECTED, txtbuf);

	if (conn_state == DEVICE_CONNECTING)
	{
		// search for ID string
		///< \note fails if header+id will be splitted across 2 or more function calls
		static const unsigned char replyHeader[] = {0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 
												0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF,
												0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF };
		static const char szId[] = "VDO/DMA 20100224";
		if (!hdr)
		{
			hdr = memsearch(data, len, replyHeader, sizeof(replyHeader));
			if (hdr)
			{
				hdr_bytes = hdr - data + sizeof(replyHeader);
				len = len - hdr_bytes;
				data += hdr_bytes;
			}
		}
		if (hdr)
		{
			unsigned char* id = memsearch(data, len, (unsigned char*)szId, strlen(szId));
			if (id)
			{
				// skip 4096 bytes starting from hdr, this is counter for skipped bytes
				hdr_bytes += id - data + strlen(szId);
				len -= id - data + strlen(szId);
				data += id - data + strlen(szId);
				SetConnectionState(DEVICE_CONNECTED, szId);
			}
		}
	}

	if (conn_state == DEVICE_CONNECTED)
	{
		if (hdr_bytes < ThreadComm::E_USB_FRAME_SIZE)
		{
			int add = std::min(len, ThreadComm::E_USB_FRAME_SIZE-hdr_bytes);
			hdr_bytes += add;
			data += add;
			len -= add;
            log("ID header skipped");
		}
		if (len == 0)
			return;

		if (frame_position == 0)
		{
			settings = upcoming_settings;
			decimator.SetFactor(settings.iSamplingPeriod);
		}

		// sign inversion - device amplifier is inverting, we have to compensate this
		for (unsigned int i=0; i<len; i++)
		{
			samples[i] = 127 - data[i];
		}

		int iSamplesCnt = decimator.Process(samples, len, output_buffer);
		if (iSamplesCnt <= 0)
			return;
		//for (int i=0; i<iSamplesCnt; i++)
		//{
		//	deque_trigger.push_back(output_buffer[i]);
		//	if (deque_trigger.size() > 2 * ThreadComm::E_USB_FRAME_SIZE)
		//		deque_trigger.pop_front();
		//}

		short* samples_p = output_buffer;

		if (frame_position == 0)
		{
			int trigger_pos = TestTrigger(iSamplesCnt);
			if (trigger_pos < 0)
			{
				// ignore this data portion
				return;
			}
			else
			{
				samples_p += trigger_pos;
				iSamplesCnt -= trigger_pos;
				log("callbackTrigger");
				callbackTrigger();
			}
		}
		//log("callbackData %d B", len * sizeof(samples[0]));
		if (iSamplesCnt + frame_position < settings.iBufferSize)
		{
			callbackData(samples_p, iSamplesCnt * sizeof(output_buffer[0]));
			frame_position += iSamplesCnt;
		}
		else
		{
			callbackData(samples_p, (settings.iBufferSize-frame_position) * sizeof(output_buffer[0]));
			frame_position += (settings.iBufferSize-frame_position);
			iSamplesCnt -= (settings.iBufferSize-frame_position);
			memmove(output_buffer, output_buffer+(settings.iBufferSize-frame_position),
				iSamplesCnt * sizeof(output_buffer[0]));
			frame_position = iSamplesCnt;
			if (settings.eTriggerType == TRIGGER_SLOPE_UP ||
				settings.eTriggerType == TRIGGER_SLOPE_DOWN ||
				settings.eTriggerType == TRIGGER_MANUAL)
			{
				frame_position = 0;
				settings = upcoming_settings;
				decimator.SetFactor(settings.iSamplingPeriod);
			}
			else if (settings.eTriggerType == TRIGGER_CONTINUOUS)
			{
				settings = upcoming_settings;
				decimator.SetFactor(settings.iSamplingPeriod);
				if (settings.eTriggerType == TRIGGER_CONTINUOUS)
				{
					callbackTrigger();
					callbackData(output_buffer, iSamplesCnt * sizeof(output_buffer[0]));
					if (frame_position == 0)
						bManualTrigger = true;
				}
				else
				{
                	frame_position = 0;
				}
			}
		}
	}
}

void ArmScope::SetConnectionState(enum E_CONNECTION_STATE state, const char* const descr)
{
	conn_state = state;
	if (callbackConnect)
		callbackConnect(state, (char*)descr);
	log("Connection state changed to %d. %s", (int)state, descr);
}

int ArmScope::TestTrigger(int len)
{
	int pos = -1;
	int level = 1.27f * (float)settings.iTriggerLevel;	//-100...100 => -127...127 range
	switch (settings.eTriggerType)
	{
	case TRIGGER_MANUAL:
		if (bManualTrigger)
			pos = 0;
		break;
	case TRIGGER_CONTINUOUS:
		if (bManualTrigger)
			pos = 0;
		break;
	case TRIGGER_SLOPE_UP:
		for (int i=0; i<len-6; i++)
		{
			if (output_buffer[i] < level && output_buffer[i+1] <= level && output_buffer[i+2] <= level &&
				output_buffer[i+3] >= level && output_buffer[i+4] >= level && output_buffer[i+5] > level)
			{
				log("i = %d, len = %d, %d %d %d %d %d %d", i, len, output_buffer[i], output_buffer[i+1], output_buffer[i+2],
					output_buffer[i+3], output_buffer[i+4], output_buffer[i+5]);
				pos = i;
				break;
			}
		}
		break;
	case TRIGGER_SLOPE_DOWN:
		for (int i=0; i<len-6; i++)
		{
			if (output_buffer[i] > level && output_buffer[i+1] >= level && output_buffer[i+2] >= level &&
				output_buffer[i+3] <= level && output_buffer[i+4] <= level && output_buffer[i+5] < level)
			{
				log("i = %d, len = %d, %d %d %d %d %d %d", i, len, output_buffer[i], output_buffer[i+1], output_buffer[i+2],
					output_buffer[i+3], output_buffer[i+4], output_buffer[i+5]);
				pos = i;
				break;
			}
		}
		break;
	default:
		break;
	}
	bManualTrigger = false;
	return pos;
}

